# SPDX-FileCopyrightText: Peter Pentchev <roam@ringlet.net>
# SPDX-License-Identifier: BSD-2-Clause
"""Check whether some programs will build or not."""

import pathlib
import sys

import click

from check_build import defs
from check_build import parse
from check_build import process
from check_build import util


@click.command(name="check-build")
@click.option(
    "-c",
    "--config",
    required=True,
    type=click.Path(exists=True, dir_okay=False, resolve_path=True, path_type=pathlib.Path),
)
@click.option("-f", "--force", is_flag=True, default=False)
@click.option("-v", "--verbose", is_flag=True, default=False)
@click.argument("programs", type=str, nargs=-1)
def main(
    config: pathlib.Path, force: bool, verbose: bool, programs: tuple[str, ...]  # noqa: FBT001
) -> None:
    """Parse command-line arguments, build things."""
    try:
        cfg = parse.load_config(
            config=config, force=force, programs=programs, logger=util.build_logger(verbose=verbose)
        )
    except defs.CBuildError as err:
        sys.exit(f"Could not parse the configuration file: {err}")
    if not cfg.selected:
        sys.exit("No programs specified")

    skipped = []
    built_ok = []

    def build_or_skip(prog: str) -> None:
        """Build a program or skip it, updating the stats lists."""
        try:
            process.build_and_test(cfg, prog)
        except defs.SkippedProgramError:
            skipped.append(prog)
        else:
            built_ok.append(prog)

    for prog in cfg.selected:
        try:
            build_or_skip(prog)
        except defs.CBuildError as err:
            sys.exit(f"Could not build the {prog} program: {err}")

    cfg.log.info(
        "%(built_ok)d program%(built_plu)s built and tested successfully, %(skipped)d skipped",
        {
            "built_ok": len(built_ok),
            "built_plu": "" if len(built_ok) == 1 else "s",
            "skipped": len(skipped),
        },
    )


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: Peter Pentchev <roam@ringlet.net>
# SPDX-License-Identifier: BSD-2-Clause
"""Parse the configuration file."""

from __future__ import annotations

import dataclasses
import functools
import sys
from typing import TYPE_CHECKING, Any

import typedload.dataloader

from check_build import defs


if sys.version_info >= (3, 11):
    import tomllib
else:
    import tomli as tomllib

if TYPE_CHECKING:
    import logging
    import pathlib
    from typing import Final


@dataclasses.dataclass
class ParseError(defs.CBuildError):
    """An error that occurred while parsing the configuration file."""


@dataclasses.dataclass
class ConfigError(ParseError):
    """An error that occurred while loading or parsing the configuration file."""

    config: pathlib.Path
    """The path to the config file that could not be parsed."""


@dataclasses.dataclass
class ConfigExcError(ConfigError):
    """An error that occurred while loading the config file with an exception attached."""

    err: Exception
    """The exception that occurred while loading the config file."""


@dataclasses.dataclass
class ConfigAbsoluteError(ConfigError):
    """The path to the config file was not an absolute one."""

    def __str__(self) -> str:
        """Provide a human-readable description of the error."""
        return f"Not an absolute path to the config file: {self.config}"


@dataclasses.dataclass
class ConfigReadError(ConfigExcError):
    """Could not read the config file."""

    def __str__(self) -> str:
        """Provide a human-readable description of the error."""
        return f"Could not read the {self.config} file: {self.err}"


@dataclasses.dataclass
class ConfigUTF8Error(ConfigExcError):
    """Could not parse the config file as valid UTF-8 text."""

    def __str__(self) -> str:
        """Provide a human-readable description of the error."""
        return f"Could not parse the {self.config} file as valid UTF-8: {self.err}"


@dataclasses.dataclass
class ConfigTOMLError(ConfigExcError):
    """Could not parse the config file as valid TOML."""

    def __str__(self) -> str:
        """Provide a human-readable description of the error."""
        return f"Could not parse the {self.config} file as valid TOML: {self.err}"


@dataclasses.dataclass
class ConfigTableError(ConfigError):
    """The config file did not contain a top-level TOML table."""

    def __str__(self) -> str:
        """Provide a human-readable description of the error."""
        return f"Expected a table as the top-level object in the {self.config} file"


@dataclasses.dataclass
class ConfigParseError(ConfigExcError):
    """Could not parse the config file as per our expected structure."""

    def __str__(self) -> str:
        """Provide a human-readable description of the error."""
        return f"Could not parse the {self.config} file: {self.err}"


@dataclasses.dataclass
class FormatParseError(ConfigExcError):
    """Could not parse the config file to obtain the format version."""

    def __str__(self) -> str:
        """Provide a human-readable description of the error."""
        return f"Could not determine the format version of the {self.config} file: {self.err}"


@dataclasses.dataclass
class FormatInvalidError(ConfigError):
    """The format version in the config file did not consist of two integers."""

    def __str__(self) -> str:
        """Provide a human-readable description of the error."""
        return f"The format.version.major/minor values in the {self.config} file must be integers"


@dataclasses.dataclass
class FormatUnsupportedError(ConfigError):
    """The format version is not supported by this version of check-build."""

    major: int
    """The major number of the format version."""

    minor: int
    """The minor number of the format version."""

    def __str__(self) -> str:
        """Provide a human-readable description of the error."""
        return f"Unsupported format version {self.major}.{self.minor} in the {self.config} file"


@dataclasses.dataclass
class UnknownProgramsError(ParseError):
    """The programs specified on the command line are not defined in the config file."""

    unknown: list[str]
    """The unknown programs specified on the command line."""

    def __str__(self) -> str:
        """Provide a human-readable description of the error."""
        return f"Unknown programs specified: {', '.join(self.unknown)}"


@dataclasses.dataclass(frozen=True)
class TCommands:
    """The commands to build and test a program."""

    clean: list[list[str]]
    build: list[list[str]]
    test: list[list[str]]


@dataclasses.dataclass(frozen=True)
class TPrereq:
    """The prerequisites for building a program."""

    programs: list[str] | None = None


@dataclasses.dataclass(frozen=True)
class TProgram:
    """The definition of a single program."""

    executable: str
    commands: TCommands
    prerequisites: TPrereq | None = None


@dataclasses.dataclass(frozen=True)
class TConfig:
    """The top-level parsed configuration settings."""

    program: dict[str, TProgram]


def _validate_format_version(config: pathlib.Path, progs: dict[str, Any]) -> None:
    """Make sure we have the appropriate format.version table."""
    try:
        major, minor = progs["format"]["version"]["major"], progs["format"]["version"]["minor"]
    except (TypeError, KeyError) as err:
        raise FormatParseError(config, err) from err
    if not isinstance(major, int) or not isinstance(minor, int):
        raise FormatInvalidError(config)
    if (major, minor) != (0, 1):
        raise FormatUnsupportedError(config, major, minor)


def _select_programs(cfg: TConfig, programs: tuple[str, ...]) -> list[str]:
    """Validate the specified list of programs or return them all."""
    if not programs:
        return list(cfg.program)

    selected: Final = list(programs)
    unknown: Final = [name for name in selected if name not in cfg.program]
    if unknown:
        raise UnknownProgramsError(unknown)

    return selected


@functools.lru_cache(maxsize=1)
def typed_loader() -> typedload.dataloader.Loader:
    """Instantiate a typed loader that can parse type annotations."""
    return typedload.dataloader.Loader(basiccast=False, failonextra=True, pep563=True)


def load_config(
    *, config: pathlib.Path, force: bool, programs: tuple[str, ...], logger: logging.Logger
) -> defs.Config:
    """Load the configuration file describing the list of programs to build."""
    if not config.is_absolute():
        raise ConfigAbsoluteError(config)
    try:
        contents: Final = config.read_text(encoding="UTF-8")
    except OSError as err:
        raise ConfigReadError(config, err) from err
    except ValueError as err:
        raise ConfigUTF8Error(config, err) from err

    try:
        progs: Final = tomllib.loads(contents)
    except ValueError as err:
        raise ConfigTOMLError(config, err) from err
    if not isinstance(progs, dict):
        raise ConfigTableError(config)

    _validate_format_version(config, progs)
    del progs["format"]
    try:
        cfg: Final = typed_loader().load(progs, TConfig)
    except ValueError as err:
        raise ConfigParseError(config, err) from err

    return defs.Config(
        force=force,
        log=logger,
        program={
            name: defs.Program(
                executable=prog.executable,
                prerequisites=defs.Prerequisites(programs=prog.prerequisites.programs)
                if prog.prerequisites is not None
                else None,
                commands=defs.Commands(
                    clean=prog.commands.clean, build=prog.commands.build, test=prog.commands.test
                ),
            )
            for name, prog in cfg.program.items()
        },
        selected=_select_programs(cfg, programs),
        topdir=config.parent,
    )



# SPDX-FileCopyrightText: Peter Pentchev <roam@ringlet.net>
# SPDX-License-Identifier: BSD-2-Clause
"""Common utilities for the check-build library."""

import functools
import logging
import sys


@functools.lru_cache(maxsize=2)
def build_logger(*, verbose: bool) -> logging.Logger:
    """Build a logger that outputs messages to the standard output stream."""
    name_prefix = "" if verbose else "non-"
    logger = logging.getLogger(f"check-build/{name_prefix}verbose")
    logger.setLevel(logging.DEBUG)

    stdout_handler = logging.StreamHandler(stream=sys.stdout)
    stdout_handler.addFilter(lambda record: record.levelno == logging.INFO)
    stdout_handler.setLevel(logging.INFO)
    logger.addHandler(stdout_handler)

    stderr_handler = logging.StreamHandler(stream=sys.stderr)
    if verbose:
        stderr_handler.addFilter(lambda record: record.levelno != logging.INFO)
        stderr_handler.setLevel(logging.DEBUG)
    else:
        stderr_handler.setLevel(logging.WARNING)
    logger.addHandler(stderr_handler)

    return logger



# SPDX-FileCopyrightText: Peter Pentchev <roam@ringlet.net>
# SPDX-License-Identifier: BSD-2-Clause
"""Build and test a single program."""

from __future__ import annotations

import dataclasses
import os
import pathlib
import shlex
import shutil
import subprocess
import tempfile
from typing import Final

from check_build import defs


@dataclasses.dataclass
class ProcessError(defs.CBuildError):
    """An error that occurred during the build and test process."""


@dataclasses.dataclass
class RunError(ProcessError):
    """Could not run a program for a build stage."""

    tag: str
    """The stage at which a program failed."""

    cmdstr: str
    """The command we attempted to run."""

    cwd: pathlib.Path
    """The directory we attempted to run the program in."""

    err: Exception
    """The error that occurred while running the program."""

    def __str__(self) -> str:
        """Provide a human-readable description of the error."""
        return (
            f"A {self.tag} command failed: could not run `{self.cmdstr}` in {self.cwd}: {self.err}"
        )


@dataclasses.dataclass
class NoProgramError(ProcessError):
    """A program was not defined in the configuration."""

    prog: str
    """The program we expected to be able to be build."""

    def __str__(self) -> str:
        """Provide a human-readable description of the error."""
        return f"No {self.prog!r} program defined in the configuration"


@dataclasses.dataclass
class CopyError(ProcessError):
    """Could not copy a source directory to a temporary one."""

    origdir: pathlib.Path
    """The original directory in the source tree."""

    srcdir: pathlib.Path
    """The temporary directory to build and run the program in."""

    err: Exception
    """The error that occurred while running the program."""

    def __str__(self) -> str:
        """Provide a human-readable description of the error."""
        return f"Could not copy {self.origdir} to the temporary {self.srcdir}: {self.err}"


@dataclasses.dataclass
class NotSourceError(ProcessError):
    """The expected directory does not contain our configuration files."""

    origdir: pathlib.Path
    """The directory that was not as expected."""

    def __str__(self) -> str:
        """Provide a human-readable description of the error."""
        return f"Not a source directory: {self.origdir}"


def _run_commands(
    cfg: defs.Config, tag: str, commands: list[list[str]], *, cwd: pathlib.Path
) -> None:
    """Run a series of commands."""
    for cmd in commands:
        cmdstr = shlex.join(cmd)
        cfg.log.debug("- running `%(cmdstr)s`", {"cmdstr": cmdstr})
        try:
            subprocess.run(cmd, check=True, cwd=cwd, shell=False)  # noqa: S603
        except (OSError, subprocess.CalledProcessError) as err:
            raise RunError(tag, cmdstr, cwd, err) from err


def _prereqs_programs_ok(
    cfg: defs.Config, prereqs: defs.Prerequisites, srcdir: pathlib.Path
) -> bool:
    """If any programs have been specified as prerequisites, look for them in the search path."""
    programs: Final = prereqs.programs
    if not programs:
        return True

    def _check_prog(cmd: str) -> bool:
        """Check for the presence of a single program."""
        prog_path: Final = pathlib.Path(cmd)
        if len(prog_path.parts) != 1 or prog_path.is_absolute():
            target: Final = srcdir / prog_path
            if not os.access(target, os.F_OK | os.X_OK):
                cfg.log.info(
                    "- the %(target)s prerequisite program is not present",
                    {"target": shlex.quote(str(target))},
                )
                return False

            return True

        if shutil.which(cmd) is None:
            cfg.log.info(
                "- the %(cmd)s prerequisite external program is not present",
                {"cmd": shlex.quote(cmd)},
            )
            return False

        return True

    cfg.log.info("Checking for programs present: %(progs)s", {"progs": shlex.join(programs)})
    return all(_check_prog(prog) for prog in programs)


def _prereqs_ok(cfg: defs.Config, prog: str, progdef: defs.Program, srcdir: pathlib.Path) -> bool:
    """If any prerequisites have been specified, check whether they are present."""
    prereqs: Final = progdef.prerequisites
    if prereqs is None:
        return True

    cfg.log.info("Checking for the prerequisites of %(prog)s", {"prog": prog})

    if not _prereqs_programs_ok(cfg, prereqs, srcdir):
        return False

    return True


def build_and_test(cfg: defs.Config, prog: str) -> None:
    """Build and test a single program in a temporary directory."""
    progdef: Final = cfg.program.get(prog)
    if progdef is None:
        raise NoProgramError(prog)

    origdir: Final = cfg.topdir / prog
    if not origdir.is_dir():
        raise NotSourceError(origdir)

    with tempfile.TemporaryDirectory(prefix="check-build.") as tempd_obj:
        tempd: Final = pathlib.Path(tempd_obj)
        srcdir: Final = tempd / prog
        cfg.log.debug("Using %(tempd)s to build %(prog)s", {"tempd": tempd, "prog": prog})

        # Copy the source files
        cfg.log.debug("Copying %(origdir)s to %(srcdir)s", {"origdir": origdir, "srcdir": srcdir})
        try:
            shutil.copytree(origdir, srcdir, symlinks=True)
        except OSError as err:
            raise CopyError(origdir, srcdir, err) from err

        if not _prereqs_ok(cfg, prog, progdef, srcdir):
            if not cfg.force:
                cfg.log.info("SKIPPING %(prog)s, prerequisites not satisfied", {"prog": prog})
                raise defs.SkippedProgramError(prog)

            cfg.log.info("NOT SKIPPING %(prog)s despite unsatisfied prerequisites", {"prog": prog})

        cfg.log.info("Building %(prog)s in %(srcdir)s", {"prog": prog, "srcdir": srcdir})
        _run_commands(cfg, "build", progdef.commands.build, cwd=srcdir)

        cfg.log.info("Testing %(prog)s in %(srcdir)s", {"prog": prog, "srcdir": srcdir})
        _run_commands(cfg, "test", progdef.commands.test, cwd=srcdir)

        cfg.log.info(
            "Successfully built and run %(prog)s in %(srcdir)s", {"prog": prog, "srcdir": srcdir}
        )
